# floor_id 是revit的楼层id


CREATE OR REPLACE FUNCTION rel_affected_space(floor_id character varying) RETURNS boolean AS
$$
from relations.src.affected_space.affected_space import get_affected_spaced
from shapely.geometry import Polygon
import json

try:
    # 获取当前楼层绑定的数据中心楼层的id
    floor_plan = plpy.prepare("SELECT id FROM public.floor where model_id = $1", ["text"])
    floor_data = floor_plan.execute([floor_id])
    floor_arr = []
    for fl_row in floor_data:
        floor_arr.append(fl_row['id'])

    if len(floor_arr) == 0:
        plpy.info("no binding floor")
        return True
    floor_str = ""
    for fl in floor_arr:
        floor_str = (floor_str + "'" + fl + "',")
    floor_str = floor_str.strip(',')
    plpy.info(floor_str)
    # 获取当前模型和前一个版本的模型id
    model_plan = plpy.prepare("select id from revit.model_file where model_floor_id = $1 and removed = false and status = 4 and version is not null order by version desc", ["text"])
    model_data = model_plan.execute([floor_id], 2)
    if len(model_data) != 2:
        plpy.info("no previous model")
        return True
    sql_str = "select rel.floor_id, rel.space_id, sp.outline from r_sp_in_fl rel left join zone_space_base sp on rel.space_id = sp.id where rel.floor_id in (" + floor_str + ")"
    # 查询出来的跟模型可能相关的所有业务空间
    space_data = plpy.execute(sql_str)
    if len(space_data) == 0:
        plpy.info("no space relation under binding floor")
        return True
    plpy.info('space data count : {0}'.format(len(space_data)))
    # 获取新模型的元空间
    new_ispace_plan = plpy.prepare("select source_id, outline from revit.space where model_id = $1", ["text"])
    new_ispace_data = new_ispace_plan.execute([model_data[0]['id']])
    plpy.info('new_ispace_data count : {0}'.format(len(new_ispace_data)))
    # 获取旧模型的所有元空间
    prev_ispace_plan = plpy.prepare("select source_id, outline from revit.space where model_id = $1", ["text"])
    prev_ispace_data = prev_ispace_plan.execute([model_data[1]['id']])
    plpy.info('prev_ispace_data count : {0}'.format(len(prev_ispace_data)))
    if len(prev_ispace_data) == 0:
        plpy.info("no prev ispace data")
        return True

    affected_space_ids = get_affected_spaced(space_data, new_ispace_data, prev_ispace_data)



    plpy.info('affected_space_ids count : {0}'.format(len(affected_space_ids)))
    for space_id in affected_space_ids:
        space_plan = plpy.prepare("update zone_space_base set state = 1 where id = $1", ["text"])
        space_plan.execute([space_id])

    return True
except Exception as e:
    plpy.info(e)
    return False
$$
LANGUAGE 'plpython3u' VOLATILE;



#
# space_outline_json_map = dict()
# result_arr = []
# # 每个楼层的每个业务空间分别和别的楼层的每个业务空间判断is_vertically_overlap
# # 将结果是true的两个业务空间保存起来
# for building_id, floor_map in row_map.items():
#     for floor_id, type_map in floor_map.items():
#         for object_type, row_arr in type_map.items():
#             # 要被对比的楼层
#             for other_floor_id in floor_map.keys():
#                 if other_floor_id == floor_id:
#                     continue
#                 other_type_map = floor_map.get(other_floor_id)
#                 if object_type not in other_type_map:
#                     continue
#                 other_row_arr = other_type_map.get(object_type)
#                 for row in row_arr:
#                     for other_row in other_row_arr:
#                         space_id = row['space_id']
#                         other_space_id = other_row['space_id']
#                         if space_id == other_space_id:
#                             continue
#                         if space_id not in space_outline_json_map:
#                             outline_json = json.loads(row['outline'])
#                             space_outline_json_map[space_id] = outline_json
#                         if other_space_id not in space_outline_json_map:
#                             other_outline_json = json.loads(other_row['outline'])
#                             space_outline_json_map[other_space_id] = other_outline_json
#                         outline = space_outline_json_map[space_id]
#                         other_outline = space_outline_json_map[other_space_id]
#                         if is_vertically_overlap(outline, other_outline):
#                             single_result = []
#                             single_result.append(space_id)
#                             single_result.append(other_space_id)
#                             single_result.append(object_type)
#                             result_arr.append(single_result)
# if len(result_arr) == 0:
#     return True
# # 删除旧业务空间的垂直交通关系(自动计算的), 添加新关系
# # 将下面对数据库的操作作为一个事务, 出异常则自动rollback
# with plpy.subtransaction():
#     del_plan = plpy.prepare("delete from r_sp_vertical_sp where project_id = $1 and sign = 2", ["text"])
#     del_plan.execute([project_id])
#     for single_result in result_arr:
#         del_manual_plan = plpy.prepare(
#             "delete from r_sp_vertical_sp where (space_id = $1 and space_other_id = $2) or (space_other_id = $1 and space_id = $2)",
#             ["text", "text"])
#         del_manual_plan.execute([single_result[0], single_result[1]])
#         insert_plan = plpy.prepare(
#             "insert into r_sp_vertical_sp(space_id, space_other_id, project_id, sign, object_type) values($1, $2, $3, 2, $4)",
#             ["text", "text", "text", "text"])
#         insert_plan.execute([single_result[0], single_result[1], project_id, single_result[2]])
#
#
#
#
#
#
#
#
#
#
# # 获取Polygon对象
# def get_polygon(single_poly):
#     poly_len = len(single_poly)
#     poly = []
#     for i in range(poly_len):
#         pair = single_poly[i]
#         poly.append((pair["X"], pair["Y"]))
#     return Polygon(poly)
#
# # 在polygon1包含polygon2的时候, 检测是否polygon1内的空洞也包含polygon2
# def is_include(polygon1, poly2):
#     length1 = len(polygon1)
#     for i in range(1, length1):
#         poly1 = get_polygon(polygon1[i])
#         if poly1.overlaps(poly2):
#             return True
#         if poly1.equals(poly2) or poly1.contains(poly2):
#             return False
#     return True
#
# def is_sub_outline_overlap(polygon1, polygon2):
#     poly1 = get_polygon(polygon1[0])
#     poly2 = get_polygon(polygon2[0])
#     if poly1.overlaps(poly2) or poly1.equals(poly2):
#         return True
#     if poly1.contains(poly2) or poly1.equals(poly2):
#         return is_include(polygon1, poly2)
#     if poly2.contains(poly1) or poly2.equals(poly1):
#         return is_include(polygon2, poly1)
#     return False
#
# # 是否垂直方向上面积有重叠
# def is_vertically_overlap(polygon1, polygon2):
#     length1 = len(polygon1)
#     length2 = len(polygon2)
#     if length1 == 0 or length2 == 0:
#         return False
#
#     for i in range(length1):
#         for j in range(length2):
#             if is_sub_outline_overlap(polygon1[i], polygon2[j]):
#                 return True
#     return False
#
# # building -> floor -> object_type -> [space_id]
# def compose_dict(zone_data):
#     building_map = dict()
#     for row in zone_data:
#         building_id = row['building_id']
#         floor_id = row['floor_id']
#         object_type = row['object_type']
#         if building_id not in building_map:
#             building_map[building_id] = dict()
#         floor_map = building_map[building_id]
#         if floor_id not in floor_map:
#             floor_map[floor_id] = dict()
#         type_map = floor_map[floor_id]
#         if object_type not in type_map:
#             type_map[object_type] = []
#         arr = type_map[object_type]
#         arr.append(row)
#     return building_map



















CREATE OR REPLACE FUNCTION "public"."rel_sp2fl"("project_id" varchar)
  RETURNS "pg_catalog"."bool" AS $BODY$
from matplotlib.path import Path
import json
from shapely.geometry import Polygon

# 获取Polygon对象
def get_polygon(single_poly):
    poly_len = len(single_poly)
    poly = []
    for i in range(poly_len):
        pair = single_poly[i]
        poly.append((pair["X"], pair["Y"]))
    return Polygon(poly)

# 判断业务空间轮廓是否跟楼层轮廓有重叠
def is_space_floor_overlap(space_outline, floor_outline):
    try:
        if space_outline is None or floor_outline is None:
            return False
        space_json = json.loads(space_outline)
        floor_json = json.loads(floor_outline)
        if len(floor_json) < 3 or len(space_json) == 0:
            return False
        floor_poly = get_polygon(floor_json)
        for i in range(len(space_json)):
            sub_space = space_json[i]
            sub_space_poly = get_polygon(sub_space[0])
            if floor_poly.contains(sub_space_poly) or floor_poly.equals(sub_space_poly) or floor_poly.overlaps(sub_space_poly) or sub_space_poly.contains(floor_poly):
                return True
        return False
    except Exception as e:
        plpy.info('计算轮廓异常')
        plpy.info(e)
        return False

column_model_id = 'model_id'
column_space_id = 'space_id'
column_floor_id1 = 'id1'
column_floor_outline1 = 'outline1'
column_floor_building1 = 'bd1'
column_floor_id2 = 'id2'
column_floor_outline2 = 'outline2'
column_floor_building2 = 'bd2'
column_space_zone = 'object_type'

# 构建一个  model_id --> set{floor_id} 的dict
def compose_model_id_floor_dict(involved_floors):
    model_id_floor_dict = dict()
    floor_outline_dict = dict()
    floor_building_dict = dict()
    for row in involved_floors:
        model_id = row.get(column_model_id)
        if model_id not in model_id_floor_dict:
            model_id_floor_dict[model_id] = set()
        floor_set = model_id_floor_dict[model_id]
        id1 = row.get(column_floor_id1)
        outline1 = row.get(column_floor_outline1)
        building_id1 = row.get(column_floor_building1)
        id2 = row.get(column_floor_id2)
        outline2 = row.get(column_floor_outline2)
        building_id2 = row.get(column_floor_building2)
        floor_set.add(id1)
        floor_set.add(id2)
        floor_building_dict[id1] = building_id1
        floor_building_dict[id2] = building_id2
        if id1 not in floor_outline_dict:
            floor_outline_dict[id1] = outline1
        if id2 not in floor_outline_dict:
            floor_outline_dict[id2] = outline2
    return model_id_floor_dict, floor_outline_dict, floor_building_dict

# 计算一个楼层组内的业务空间与楼层的轮廓关系
def calc_floor_group(floor_set, floor_outline_dict):
    sql_str = ""
    for floor_id in floor_set:
        sql_str += '\'' + floor_id + '\','
    if sql_str.endswith(','):
        sql_str = sql_str[0:-1]
    plpy.info('sql : {0}'.format(sql_str))
    involved_space_plan = plpy.prepare("SELECT distinct rel.space_id space_id, sp.outline outline1, sp.object_type object_type " +
        "FROM r_sp_in_fl rel inner join zone_space_base sp on sp.id = rel.space_id where rel.floor_id in ({0}) and sp.outline is not null".format(sql_str), ["text"])
    involved_space = involved_space_plan.execute([project_id])
    plpy.info("involved space:{}".format(len(involved_space)))
    result_dict = dict()    # space_id --> set{floor_id}
    space_zone_dict = dict()   # space_id --> object_type
    for space_row in involved_space:
        space_outline = space_row.get(column_floor_outline1)
        space_id = space_row.get(column_space_id)
        space_zone = space_row.get(column_space_zone)
        space_zone_dict[space_id] = space_zone
        result_dict[space_id] = set()
        rel_floor_set = result_dict[space_id]
        for floor_id in floor_set:
        #for floor_id, floor_outline in floor_outline_dict.items():
            floor_outline = floor_outline_dict[floor_id]
            if is_space_floor_overlap(space_outline, floor_outline):
                plpy.info('{0} : {1}'.format(space_id, floor_id))
                rel_floor_set.add(floor_id)
    return result_dict, space_zone_dict

# 删除以前的关系
def delete_prev_rel(floor_set):
    sql_str = ''
    for floor_id in floor_set:
        sql_str += '\'{0}\','.format(floor_id)
    if sql_str.endswith(','):
        sql_str = sql_str[0:-1]
    delete_rel_plan = plpy.prepare("delete from public.r_sp_in_fl where floor_id in ({0})".format(sql_str), [])
    delete_rel_plan.execute([])

# 添加关系到数据库
def add_rel(result_dict, space_zone_dict, floor_building_dict):
    for space_id, floor_set in result_dict.items():
        space_zone = space_zone_dict[space_id]
        for floor_id in floor_set:
            floor_building = floor_building_dict[floor_id]
            plpy.info("space_id : {0}, floor_id : {1}".format(space_id, floor_id))
            insert_rel_plan = plpy.prepare("insert into public.r_sp_in_fl(floor_id, space_id, object_type, project_id, sign, building_id) " +
                "values($1, $2, $3, $4, 2, $5)", ["text", "text", "text", "text", "text"])
            try:
                insert_rel_plan.execute([floor_id, space_id, space_zone, project_id, floor_building])
            except Exception as e:
                plpy.info(e)

try:
    # 将下面对数据库的操作作为一个事务, 出异常则自动rollback

    involved_floors_plan = plpy.prepare("select f1.id id1, f1.building_id bd1, f2.id id2, f2.building_id bd2, f1.model_id model_id, f1.outline outline1, f2.outline outline2 " +
        "from floor f1 inner join floor f2 on f1.model_id = f2.model_id " +
        "where f1.id != f2.id and f1.outline is not null and f2.outline is not null and f1.project_id = $1 and f2.project_id = $1", ["text"])
    involved_floors = involved_floors_plan.execute([project_id])
    if len(involved_floors) == 0 :
        plpy.info('没有需要计算的改变')
        return True
    plpy.info(len(involved_floors))
    model_id_floor_dict, floor_outline_dict, floor_building_dict = compose_model_id_floor_dict(involved_floors)
    plpy.info(len(floor_outline_dict))
    plpy.info(len(floor_building_dict))
    plpy.info(len(model_id_floor_dict))
    for model_id, floor_set in model_id_floor_dict.items():
        with plpy.subtransaction():
            plpy.info(model_id)
            plpy.info(floor_set)
            result_dict, space_zone_dict = calc_floor_group(floor_set, floor_outline_dict)  # space_id -->所属 set{floor_id},  space_id --> object_type
            delete_prev_rel(floor_set)
            add_rel(result_dict, space_zone_dict, floor_building_dict)
    return True
except Exception as e:
    plpy.warning(e)
    return False
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100;




















CREATE OR REPLACE FUNCTION "public"."rel_eq2sp"("tables" text, "out_tables" text, "project_id" varchar, "sign1" int4, "sign2" int4)
  RETURNS "pg_catalog"."bool" AS $BODY$
from matplotlib.path import Path
import json

def is_in_meta_polygon(point, single_poly, radius):
    poly_len = len(single_poly)
    poly = []
    for i in range(poly_len):
        pair = single_poly[i]
        poly.append((pair["X"], pair["Y"]))
    p = Path(poly)
    return p.contains_points([point], None, radius)


def is_in_polygon(point, polygons):
    polygons_length = len(polygons)
    if polygons_length == 0:
        return False
    for j in range(polygons_length):
        polygon = polygons[j]
        if j == 0:
            if not is_in_meta_polygon(point, polygon, -0.001):
                return False
        else:
            if is_in_meta_polygon(point, polygon, 0.001):
                return False
    return True


def is_point_in_polygon(x, y, json_poly):
    try:
        polygon_list = json.loads(json_poly)
        total_len = len(polygon_list)
        point_pair = (float(x), float(y))
        for index in range(total_len):
            if is_in_polygon(point_pair, polygon_list[index]):
                return True
        return False
    except Exception as e:
        plpy.info(e)
        return False


# 将下面对数据库的操作作为一个事务, 出异常则自动rollback
input_table_list = tables.split(',')
output_table_list = out_tables.split(',')
with plpy.subtransaction():
    for i in range(0, len(input_table_list)):
        in_table_name = input_table_list[i]
        out_table_name = output_table_list[i]
        # 删除原来关系表中的数据
        plan1 = plpy.prepare("delete from {0} where project_id = $1 and (sign = $2 or sign = $3)".format(out_table_name.strip()), ["text", "integer", "integer"])
        plan1.execute([project_id, sign1, sign2])
        # 计算关系
        plan_floor = plpy.prepare("select id from floor where project_id = $1", ["text"])
        floors = plan_floor.execute([project_id])
        # 按楼层计算
        for floor in floors:
            floor_id = floor['id']
            # 获取楼层下的设备
            plan_equip = plpy.prepare("select id, bim_location from equipment where project_id = $1 and bim_location is not null and floor_id = $2", ["text", "text"])
            equips = plan_equip.execute([project_id, floor_id])
            if len(equips) == 0:
                continue
            # 获取楼层下的业务空间

            space_plan = plpy.prepare("select sp.id, sp.outline, sp.object_type from public.r_sp_in_fl rel inner join {0} sp on rel.space_id = sp.id where outline is not null and rel.floor_id = $1".format(in_table_name), ["text"])
            spaces = space_plan.execute([floor_id])
            if len(spaces) == 0:
                continue
            # 判断设备的bim_location是否在业务空间的outline内
            for equip in equips:
                for space in spaces:
                    try:
                        location = equip['bim_location'].split(',')
                        if is_point_in_polygon(location[0], location[1], space['outline']):
                            # 设备在业务空间内, 添加关系
                            insert_plan = plpy.prepare("insert into {0}(eq_id, sp_id, project_id, sign, type, zone_type) values($1, $2, $3, 2, 'eq2sp_in', $4)".format(out_table_name.strip()), ["text", "text", "text", "text"])
                            insert_plan.execute([equip['id'], space['id'], project_id, space['object_type']])
                    except Exception as ex:
                        plpy.warning(ex)
                        continue
return True
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100


























CREATE OR REPLACE FUNCTION "revit_calc"."check_grid_valid"("project_id" varchar, "folder_id" varchar, "model_id" varchar)
  RETURNS TABLE("id" varchar, "floor_name" varchar, "project_id" varchar, "folder_id" varchar, "fid" varchar, "accept_time" timestamptz, "version" int4, "note" text, "user_id" varchar, "user_name" varchar, "log" jsonb, "url" text, "md5" varchar, "status" int2, "group_id" int4, "plan_name" varchar) AS $BODY$
    from relations.src.grid.check_grid import check_grid_upright


    involved_model_sql = "select * from revit.model_floor_file_func('" + project_id + "') where folder_id = '"+ folder_id +"' and ( status in (4) or  fid = '" + model_id + "')"
    grid_sql = "select * from revit.grid where model_id = "

    involved_model = plpy.execute(involved_model_sql)
    grid_data = dict()
    for item in involved_model:
        current_grid_sql = grid_sql + '\'{model_id}\''.format(model_id=item.get('fid'))
        single_model_grid = plpy.execute(current_grid_sql)
        grid_data[item.get('fid')] = single_model_grid
    result = check_grid_upright(involved_model, grid_data)
    return involved_model
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100
  ROWS 1000
















CREATE OR REPLACE FUNCTION "public"."is_vertically_overlap"("project_id" varchar)
  RETURNS "pg_catalog"."bool" AS $BODY$

from shapely.geometry import Polygon
import json

# 获取Polygon对象
def get_polygon(single_poly):
    poly_len = len(single_poly)
    poly = []
    for i in range(poly_len):
        pair = single_poly[i]
        poly.append((pair["X"], pair["Y"]))
    return Polygon(poly)

# 在polygon1包含polygon2的时候, 检测是否polygon1内的空洞也包含polygon2
def is_include(polygon1, poly2):
    length1 = len(polygon1)
    for i in range(1, length1):
        poly1 = get_polygon(polygon1[i])
        try:
            if poly1.overlaps(poly2):
                return True
        except Exception as e:
            return False
        if poly1.equals(poly2) or poly1.contains(poly2):
            return False
    return True

def is_sub_outline_overlap(polygon1, polygon2):
    try:
        poly1 = get_polygon(polygon1[0])
        poly2 = get_polygon(polygon2[0])
        if poly1.overlaps(poly2) or poly1.equals(poly2):
            return True
        if poly1.contains(poly2):
            return is_include(polygon1, poly2)
        if poly2.contains(poly1):
            return is_include(polygon2, poly1)
    except Exception as e:
        return False
    return False

# 是否垂直方向上面积有重叠
def is_vertically_overlap(polygon1, polygon2):
    length1 = len(polygon1)
    length2 = len(polygon2)
    if length1 == 0 or length2 == 0:
        return False

    for i in range(length1):
        for j in range(length2):
            if is_sub_outline_overlap(polygon1[i], polygon2[j]):
                return True
    return False


# building -> floor -> object_type -> [space_id]
def compose_dict(zone_data):
    building_map = dict()
    for row in zone_data:
        building_id = row['building_id']
        floor_id = row['floor_id']
        object_type = row['object_type']
        if building_id not in building_map:
            building_map[building_id] = dict()
        floor_map = building_map[building_id]
        if floor_id not in floor_map:
            floor_map[floor_id] = dict()
        type_map = floor_map[floor_id]
        if object_type not in type_map:
            type_map[object_type] = []
        arr = type_map[object_type]
        arr.append(row)
    return building_map


try:
    # 获取所有建筑, for循环获取每个建筑下所有的业务空间, 按楼层分类
    zone_plan = plpy.prepare("SELECT rel.space_id, fl.building_id, rel.floor_id, rel.object_type, sp.outline FROM r_sp_in_fl rel LEFT JOIN public.floor fl on fl.id = rel.floor_id left join zone_space_base sp on rel.space_id = sp.id where rel.project_id = $1 and sp.outline is not null", ["text"])
    zone_data = zone_plan.execute([project_id])
    if len(zone_data) <2:
        return True
    row_map = compose_dict(zone_data)

    space_outline_json_map = dict()
    result_arr = []
    # 每个楼层的每个业务空间分别和别的楼层的每个业务空间判断is_vertically_overlap
    # 将结果是true的两个业务空间保存起来
    for building_id, floor_map in row_map.items():
        checked_floor = set()
        for floor_id, type_map in floor_map.items():
            checked_floor.add(floor_id)
            for object_type, row_arr in type_map.items():
                # 要被对比的楼层
                for other_floor_id in floor_map.keys():
                    if other_floor_id in checked_floor:
                        continue
                    other_type_map = floor_map.get(other_floor_id)
                    if object_type not in other_type_map:
                        continue
                    other_row_arr = other_type_map.get(object_type)
                    for row in row_arr:
                        for other_row in other_row_arr:
                            space_id = row['space_id']
                            other_space_id = other_row['space_id']
                            if space_id == other_space_id:
                                continue
                            if space_id not in space_outline_json_map:
                                outline_json = json.loads(row['outline'])
                                space_outline_json_map[space_id] = outline_json
                            if other_space_id not in space_outline_json_map:
                                other_outline_json = json.loads(other_row['outline'])
                                space_outline_json_map[other_space_id] = other_outline_json
                            outline = space_outline_json_map[space_id]
                            other_outline = space_outline_json_map[other_space_id]
                            if is_vertically_overlap(outline, other_outline):
                                single_result = []
                                single_result.append(space_id)
                                single_result.append(other_space_id)
                                single_result.append(object_type)
                                result_arr.append(single_result)
    if len(result_arr) == 0:
        return True
    plpy.info(len(result_arr))
    # 删除旧业务空间的垂直交通关系(自动计算的), 添加新关系
    # 将下面对数据库的操作作为一个事务, 出异常则自动rollback
    with plpy.subtransaction():
        del_plan = plpy.prepare("delete from r_sp_vertical_sp where project_id = $1 and sign = 2", ["text"])
        del_plan.execute([project_id])
        for single_result in result_arr:
            del_manual_plan = plpy.prepare("delete from r_sp_vertical_sp where (space_id = $1 and space_other_id = $2) or (space_other_id = $1 and space_id = $2)", ["text", "text"])
            del_manual_plan.execute([single_result[0], single_result[1]])
            insert_plan = plpy.prepare("insert into r_sp_vertical_sp(space_id, space_other_id, project_id, sign, object_type) values($1, $2, $3, 2, $4)", ["text", "text", "text", "text"])
            insert_plan.execute([single_result[0], single_result[1], project_id, single_result[2]])
    return True
except Exception as e:
    plpy.info(e)
    return False
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100


















CREATE OR REPLACE FUNCTION "public"."f_lock"("func_name" varchar, "project_id" varchar)
  RETURNS "pg_catalog"."bool" AS $BODY$

lock_plan = plpy.prepare("select func_name, func_param, reentry from public.func_lock where func_name = $1 and func_param = $2", ["text", "text"])
lock = lock_plan.execute([func_name, project_id])
if len(lock) == 0:
    try:
        insert_lock_plan = plpy.prepare("insert into public.func_lock(func_name, func_param) values($1, $2)",["text", "text"])
        lock = insert_lock_plan.execute([func_name, project_id])
    except Exception as e:
        plpy.info("insert lock error : {0}".format(e))
retry_times = 10
while retry_times > 0:
    retry_times = retry_times - 1
    lock_plan = plpy.prepare("select func_name, func_param, reentry from public.func_lock where func_name = $1 and func_param = $2", ["text", "text"])
    lock = lock_plan.execute([func_name, project_id])
    if len(lock) == 1:
        reentry_count = lock[0].get('reentry')
        if reentry_count > 0:
            update_lock_plan = plpy.prepare("update public.func_lock set reentry = $1 where func_name = $2 and func_param = $3 and reentry = $4", ["integer", "text", "text", "integer"])
            result = update_lock_plan.execute([reentry_count - 1, func_name, project_id, reentry_count])
            if result.nrows() > 0:
                return True
return False
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100









CREATE OR REPLACE FUNCTION "public"."f_unlock"("func_name" varchar, "project_id" varchar)
  RETURNS "pg_catalog"."bool" AS $BODY$

retry_times = 10
while retry_times > 0:
    retry_times = retry_times - 1
    lock_plan = plpy.prepare("select func_name, func_param, reentry from public.func_lock where func_name = $1 and func_param = $2", ["text", "text"])
    lock = lock_plan.execute([func_name, project_id])
    if len(lock) == 0:
        return True

    reentry_count = lock[0].get('reentry')
    reentry_total = lock[0].get('reentry_total')
    plpy.info(reentry_count)
    plpy.info(reentry_total)
    if reentry_count >= reentry_total :
        return True
    update_lock_plan = plpy.prepare("update public.func_lock set reentry = $1 where func_name = $2 and func_param = $3 and reentry = $4", ["integer", "text", "text", "integer"])
    result = update_lock_plan.execute([reentry_count + 1, func_name, project_id, reentry_count])
    if result.nrows() > 0:
        return True
return False
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100














CREATE OR REPLACE FUNCTION public.is_vertically_overlap(project_id varchar, _building_id varchar, zone_type varchar, space_ids _varchar)
  RETURNS "pg_catalog"."bool" AS $BODY$

from shapely.geometry import Polygon
import json

# 获取Polygon对象
def get_polygon(single_poly):
    poly_len = len(single_poly)
    poly = []
    for i in range(poly_len):
        pair = single_poly[i]
        poly.append((pair["X"], pair["Y"]))
    return Polygon(poly)

# 在polygon1包含polygon2的时候, 检测是否polygon1内的空洞也包含polygon2
def is_include(polygon1, poly2):
    length1 = len(polygon1)
    for i in range(1, length1):
        poly1 = get_polygon(polygon1[i])
        try:
            if poly1.overlaps(poly2):
                return True
        except Exception as e:
            return False
        if poly1.equals(poly2) or poly1.contains(poly2):
            return False
    return True

def is_sub_outline_overlap(polygon1, polygon2):
    try:
        poly1 = get_polygon(polygon1[0])
        poly2 = get_polygon(polygon2[0])
        if poly1.overlaps(poly2) or poly1.equals(poly2):
            return True
        if poly1.contains(poly2):
            return is_include(polygon1, poly2)
        if poly2.contains(poly1):
            return is_include(polygon2, poly1)
    except Exception as e:
        return False
    return False

# 是否垂直方向上面积有重叠
def is_vertically_overlap(polygon1, polygon2):
    length1 = len(polygon1)
    length2 = len(polygon2)
    if length1 == 0 or length2 == 0:
        return False

    for i in range(length1):
        for j in range(length2):
            if is_sub_outline_overlap(polygon1[i], polygon2[j]):
                return True
    return False


# building_map 结构  building -> floor -> object_type -> [space_id]
def compose_dict(zone_data):
    building_map = dict()
    for row in zone_data:
        building_id = row['building_id']
        floor_id = row['floor_id']
        object_type = row['object_type']
        if building_id not in building_map:
            building_map[building_id] = dict()
        floor_map = building_map[building_id]
        if floor_id not in floor_map:
            floor_map[floor_id] = dict()
        type_map = floor_map[floor_id]
        if object_type not in type_map:
            type_map[object_type] = []
        arr = type_map[object_type]
        arr.append(row)
    return building_map

def build_sql_str(space_ids_set):
    sql_str = ""
    for space_id in space_ids_set:
        sql_str += '\'' + space_id + '\','
    if sql_str.endswith(','):
        sql_str = sql_str[0:-1]
    return sql_str

try:
    if len(space_ids) < 1:
        plpy.info("no input space_ids")
        return True
    # 获取建筑下所有的业务空间, 按楼层分类
    zone_plan = plpy.prepare("SELECT rel.space_id, fl.building_id, rel.floor_id, rel.object_type, sp.outline FROM r_sp_in_fl rel LEFT JOIN public.floor fl on fl.id = rel.floor_id " +
         "left join zone_space_base sp on rel.space_id = sp.id where rel.project_id = $1 and sp.outline is not null and fl.building_id = $2 and sp.object_type = $3", ["text", "text", "text"])
    zone_data = zone_plan.execute([project_id, _building_id, zone_type])
    plpy.info('all space size : {0}'.format(len(zone_data)))
    if len(zone_data) <2:
        plpy.info("no enough space to calc")
        return True
    # 转换space_ids为set类型
    space_ids_set = set(space_ids)
    row_map = compose_dict(zone_data)

    space_outline_json_map = dict()
    result_arr = []
    # 每个楼层的每个业务空间分别和别的楼层的每个业务空间判断is_vertically_overlap
    # 将结果是true的两个业务空间保存起来
    for building_id, floor_map in row_map.items():
        checked_floor = set()
        for floor_id, type_map in floor_map.items():
            checked_floor.add(floor_id)
            for object_type, row_arr in type_map.items():
                # 要被对比的楼层
                for other_floor_id in floor_map.keys():
                    if other_floor_id in checked_floor:
                        continue
                    other_type_map = floor_map.get(other_floor_id)
                    if object_type not in other_type_map:
                        continue
                    other_row_arr = other_type_map.get(object_type)
                    for row in row_arr:
                        for other_row in other_row_arr:
                            space_id = row['space_id']
                            other_space_id = other_row['space_id']
                            if space_id not in space_ids_set and other_space_id not in space_ids_set:
                                continue
                            if space_id == other_space_id:
                                continue
                            if space_id not in space_outline_json_map:
                                outline_json = json.loads(row['outline'])
                                space_outline_json_map[space_id] = outline_json
                            if other_space_id not in space_outline_json_map:
                                other_outline_json = json.loads(other_row['outline'])
                                space_outline_json_map[other_space_id] = other_outline_json
                            outline = space_outline_json_map[space_id]
                            other_outline = space_outline_json_map[other_space_id]
                            if is_vertically_overlap(outline, other_outline):
                                single_result = []
                                single_result.append(space_id)
                                single_result.append(other_space_id)
                                single_result.append(object_type)
                                result_arr.append(single_result)
    # if len(result_arr) == 0:
    #     return True
    # 删除旧业务空间的垂直交通关系(自动计算的), 添加新关系
    # 将下面对数据库的操作作为一个事务, 出异常则自动rollback
    with plpy.subtransaction():
        sql_str = build_sql_str(space_ids_set)
        del_plan = plpy.prepare("delete from r_sp_vertical_sp where space_id in ({0}) or space_other_id in ({0})".format(sql_str), [])
        del_plan.execute([])
        for single_result in result_arr:
            #del_manual_plan = plpy.prepare("delete from r_sp_vertical_sp where (space_id = $1 and space_other_id = $2) or (space_other_id = $1 and space_id = $2)", ["text", "text"])
            #del_manual_plan.execute([single_result[0], single_result[1]])
            insert_plan = plpy.prepare("insert into r_sp_vertical_sp(space_id, space_other_id, project_id, sign, object_type) values($1, $2, $3, 2, $4)", ["text", "text", "text", "text"])
            #plpy.info(single_result)
            insert_plan.execute([single_result[0], single_result[1], project_id, single_result[2]])
    return True
except Exception as e:
    plpy.info(e)
    return False

$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100

























CREATE OR REPLACE FUNCTION "public"."sys_block"("project_id" varchar, "building_id" varchar, "system_name" varchar, "domain" varchar)
RETURNS "pg_catalog"."bool"
AS $BODY$

from relations.src.system_relation import systemutils, graph_model

# 获取楼层上的ModelId信息点
def get_models(project_id, building_id):
    sql = "select " \
          "id,name,local_id,local_name,sequence_id,model_id,building_id,project_id " \
          "from floor " \
          "where project_id=$1 " \
          "and ($2 is null or building_id=$2) " \
          "and model_id is not null " \
          "order by sequence_id desc"
    join_plan = plpy.prepare(sql, ["text", "text"])
    data = join_plan.execute([project_id, building_id])
    models = systemutils.sqldata2objlist(data)
    return models


def get_connectors(model_id, system_name, domain):
    sql = "SELECT * FROM revit.connector where model_id=$1 and mep_system_type=$2 and domain=$3"
    join_plan = plpy.prepare(sql, ["text", "text", "text"])
    data = join_plan.execute([model_id, system_name, domain])
    floor_connectors = systemutils.sqldata2objlist(data)
    return floor_connectors


def get_elements(model_id):
    floor_elements = []
    sql = "select * from revit.equipment where model_id=$1"
    join_plan = plpy.prepare(sql, ["text"])
    data = join_plan.execute([model_id])
    floor_elements.extend(systemutils.sqldata2objlist(data))

    sql = "select * from revit.component where model_id=$1"
    join_plan = plpy.prepare(sql, ["text"])
    data = join_plan.execute([model_id])
    floor_elements.extend(systemutils.sqldata2objlist(data))

    sql = "select * from revit.other where model_id=$1"
    join_plan = plpy.prepare(sql, ["text"])
    data = join_plan.execute([model_id])
    floor_elements.extend(systemutils.sqldata2objlist(data))

    sql = "select * from revit.join_object where model_id=$1"
    join_plan = plpy.prepare(sql, ["text"])
    data = join_plan.execute([model_id])
    floor_elements.extend(systemutils.sqldata2objlist(data))

    sql = "select * from revit.pipe where model_id=$1"
    join_plan = plpy.prepare(sql, ["text"])
    data = join_plan.execute([model_id])
    floor_elements.extend(systemutils.sqldata2objlist(data))

    sql = "select * from revit.duct where model_id=$1"
    join_plan = plpy.prepare(sql, ["text"])
    data = join_plan.execute([model_id])
    floor_elements.extend(systemutils.sqldata2objlist(data))
    return floor_elements


def delete_block_data(project_id, building_id, system_name, domain):
    sql = "delete from revit_calc.connected_block " \
          "where project_id=$1" \
          "and mep_system_type=$2" \
          "and domain=$3" \
          "and ($4 is null or building_id=$4)"
    join_plan = plpy.prepare(sql, ["text", "text", "text", "text"])
    join_plan.execute([project_id, system_name, domain, building_id])


def save_edge_data(edge, block_id, system_name, domain, building_id):
    vertex1 = edge.start_vertex
    vertex2 = edge.end_vertex
    c1 = vertex1.system_data
    c2 = vertex2.system_data
    floor1 = vertex1.get_parent_floor()
    floor2 = vertex2.get_parent_floor()
    model1 = floor1.model
    model2 = floor2.model
    sql = "insert into revit_calc.connected_block" \
          "(id1,type1,model_id1,id2,type2,model_id2,block_id,project_id,mep_system_type,domain,building_id) values " \
          "($1,$2,$3,$4,$5,$6,$7,$8,$9,$10,$11)"
    join_plan = plpy.prepare(sql,
                             ["text", "text", "text", "text", "text", "text", "text", "text", "text", "text", "text"])
    join_plan.execute([c1.id, c1.type, model1.model_id, c2.id, c2.type, model2.model_id, block_id, model1.project_id,
                       floor1.system_name, floor1.domain, building_id])


def get_real_model_id(models):
    sql_str = ""
    for model in models:
        sql_str += '\'' + model.model_id + '\','
    if sql_str.endswith(','):
        sql_str = sql_str[0:-1]
    real_model_id_plan = plpy.prepare("select mid.id, file.id fid, file.status from (select id, current_model_id from revit.model_floor where id in ({0})) mid left join revit.model_file file on mid.current_model_id = file.id".format(sql_str), [])
    real_model_ids = real_model_id_plan.execute([])
    floor_model_dict = dict()
    for row in real_model_ids:
        floor_id = row.get('id')
        model_id = row.get('fid')
        status = row.get('status')
        if status == 4:
            floor_model_dict[floor_id] = model_id
    for model in models:
        if model.model_id not in floor_model_dict:
            models.remove(model)
        else:
            model.model_id = floor_model_dict[model.model_id]


try:
    # 将下面对数据库的操作作为一个事务, 出异常则自动rollback
    with plpy.subtransaction():
        models = get_models(project_id, building_id)
        floorgraphs = []
        plpy.info("models : {0}".format(len(models)))
        get_real_model_id(models)
        for model in models:
            floor_connectors = get_connectors(model.model_id, system_name, domain)
            plpy.info("floor_connectors : {0}".format(len(floor_connectors)))
            floor_elements = get_elements(model.model_id)
            plpy.info("floor_elements : {0}".format(len(floor_elements)))
            g = graph_model.FloorGraph(model, system_name, domain, floor_connectors, floor_elements)
            plpy.info("reached")
            g.get_floor_graphs()
            g.combine_single_model()
            floorgraphs.append(g)
        projectgraph = graph_model.combine_systemgraph(project_id, floorgraphs)

        delete_block_data(project_id, building_id, system_name, domain)
        block_id = 0
        for g in projectgraph.groupedsystemgraphs:
            for s in g.systemgraphs:
                for e in s.system_edges:
                    save_edge_data(e, block_id, system_name, domain, building_id)
            for e in g.connectedges:
                save_edge_data(e, block_id, system_name, domain, building_id)
            block_id += 1
    return True
except Exception as e:
    plpy.warning(e)
    return False
$BODY$
LANGUAGE plpython3u VOLATILE
COST 100




















CREATE OR REPLACE FUNCTION "public"."sys_direction"("project_id" varchar, "building_id" varchar, "block_id" varchar, "system_name" varchar, "domain" varchar, "is_source" bool=true)
  RETURNS "pg_catalog"."bool" AS $BODY$
from relations.src.system_relation import calc_flowdirection, systemdatautils,systemutils

def get_connected_block_data(project_id,building_id,block_id,system_name,domain):
  sql ="select * from revit_calc.connected_block " \
		  "where project_id=$1" \
		  "and block_id=$2" \
		  "and mep_system_type=$3" \
		  "and domain=$4" \
		  "and ($5 is null or building_id=$5)" \
		  "order by cast(depth as int) asc"
  join_plan=plpy.prepare(sql,["text","text","text","text","text"])
  data=join_plan.execute([project_id,block_id,system_name,domain,building_id])
  objs=systemutils.sqldata2objlist(data)
  return objs

def get_connected_block_source_data(project_id,building_id,block_id,system_name,domain,is_source):
  sql = "select * from revit_calc.connected_block_source " \
		  "where project_id=$1" \
		  "and block_id=$2" \
		  "and mep_system_type=$3" \
		  "and domain=$4" \
		  "and ($5 is null or building_id=$5)" \
		  "and is_source=$6"
  join_plan=plpy.prepare(sql,["text","text","text","text","text","boolean"])
  data=join_plan.execute([project_id,block_id,system_name,domain,building_id,is_source])
  objs=systemutils.sqldata2objlist(data)
  return objs

def update_connected_block_data(block_data):
  direction=block_data.direction
  depth=block_data.depth
  id=block_data.id
  sql="update revit_calc.connected_block set direction=$1,depth=$2 where id=$3"
  plpy.info(sql+str(direction)+str(depth)+str(id))
  join_plan=plpy.prepare(sql,['int','int','int'])
  join_plan.execute([direction,depth,id])

try:
  # 将下面对数据库的操作作为一个事务, 出异常则自动rollback
  with plpy.subtransaction():
    block_datas=get_connected_block_data(project_id,building_id, block_id,system_name,domain)
    source_datas=get_connected_block_source_data(project_id,building_id, block_id,system_name,domain,is_source)
    block_datas= calc_flowdirection.calc(block_datas, source_datas)
    for block_data in block_datas:
      update_connected_block_data(block_data)
except Exception as e:
    plpy.warning(e)
    return False
else:
    return True
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100













CREATE OR REPLACE FUNCTION "pointconfig"."rel_description_param_key"("word" _varchar)
  RETURNS "pg_catalog"."text" AS $BODY$

import json
from kwextraction.extraction import extractor


def get_key_words(text_list):
    return extractor(text_list)

try:
    arr = list()
    for item in get_key_words(word):
        item.pop('str')
        arr.append(item)
    return json.dumps(arr, sort_keys=True, indent=0, ensure_ascii=False)
except Exception as e:
    plpy.warning(e)
    return "error"
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100





















CREATE OR REPLACE FUNCTION "public"."rel_sp2sp_v2"("project_id" varchar)
  RETURNS "pg_catalog"."bool" AS $BODY$

import math

column_project_id = 'project_id'
column_location_one = 'location_one'
column_location_two = 'location_two'
column_space_id_one = 'space_id_one'
column_space_id_two = 'space_id_two'
column_zone_type = 'zone_type'
column_added_polygon = 'polygon'

column_id = 'id'
column_bim_location = 'bim_location'
column_floor_id = 'floor_id'
column_object_type = 'object_type'
column_outline = 'outline'
key_x = 'X'
key_y = 'Y'

delta_distance = 200
door_length = 1500


# 获取两点之间的距离
def get_segment_distance(x1, y1, x2, y2):
    x_diff = x1 - x2
    y_diff = y1 - y2
    return math.sqrt(x_diff ** 2 + y_diff ** 2)


# 获取垂直方向的斜率
def get_vertical_k(k):
    if k is None:
        return 0
    if k == 0:
        return None
    return 1 / k

# 根据点 x, y 计算 y = kx + a 中的 a
def get_a_by_point_k(x, y, k):
    if k is None:
        return None
    return y - k * x


# 在直线   y = kx + a 上找到距离点 base_x, base_y 距离为distance 的坐标点(两个坐标点)
def get_point_by_distance(base_x, base_y, k, a, distance):
    if k is None:
        return base_x, base_y + distance, base_x, base_y - distance
    vector_x1 = math.sqrt(distance ** 2 / (1 + k ** 2))
    vector_x2 = -vector_x1
    vector_y1 = k * vector_x1
    vector_y2 = k * vector_x2
    return base_x + vector_x1, base_y + vector_y1, base_x + vector_x2, base_y + vector_y2

# 在直线 y = kx + a 上找到一个点, 该点距离点(base_x, base_y) 长度为 distance, 并且距离vec_x, vec_y最近
def get_point_by_distance_on_segment(base_x, base_y, k, a, distance, vec_x, vec_y):
    x1, y1, x2, y2 = get_point_by_distance(base_x, base_y, k, a, distance)
    distance1 = get_segment_distance(x1, y1, vec_x, vec_y)
    distance2 = get_segment_distance(x2, y2, vec_x, vec_y)
    if distance1 > distance2:
        return x2, y2
    return x1, y1


# 获取输入的两点之间开门之后门的坐标点
# 返回格式
# 如果线段距离小于等于door_length的情况 : x1, y1, x2, y2
# 如果线段距离大于door_length的情况 (两端各开一个门): x1, y1, x2, y2, x3, y3, x4, y4
def get_points(x1, y1, x2, y2):
    # 如果两个点是同一个点
    if x1 == y1 and x2 == y2:
        return None, None, None, None
    # 计算线段的距离
    distance = get_segment_distance(x1, y1, x2, y2)
    # 计算当前线段的 k
    if x1 == x2:
        k = None
    else:
        k = (y1 - y2) / (x1 - x2)
    # 计算垂直方向的k
    vertical_k = get_vertical_k(k)
    # 计算当前线段的 a
    a = get_a_by_point_k(x1, y1, k)
    # 距离大于delta_distance, 则足够开两个门
    if distance > delta_distance:
        seg_x1, seg_y1 = get_point_by_distance_on_segment(x1, y1, k, a, delta_distance, x2, y2)
        seg_x2, seg_y2 = get_point_by_distance_on_segment(x2, y2, k, a, delta_distance, x1, y1)
        vertical_a1 = get_a_by_point_k(seg_x1, seg_y1, vertical_k)
        vertical_a2 = get_a_by_point_k(seg_x2, seg_y2, vertical_k)
        dest_x1, dest_y1, dest_x2, dest_y2 = get_point_by_distance(seg_x1, seg_y1, vertical_k, vertical_a1, door_length)
        dest_x3, dest_y3, dest_x4, dest_y4 = get_point_by_distance(seg_x2, seg_y2, vertical_k, vertical_a2, door_length)
        return [dest_x1, dest_x2, dest_x3, dest_x4], [dest_y1, dest_y2, dest_y3, dest_y4]
    else:
        # 距离太小, 在中间开门
        seg_x1, seg_y1 = get_point_by_distance_on_segment(x1, y1, k, a, distance/2, x2, y2)
        vertical_a1 = get_a_by_point_k(seg_x1, seg_y1, vertical_k)
        dest_x1, dest_y1, dest_x2, dest_y2 = get_point_by_distance(seg_x1, seg_y1, vertical_k, vertical_a1, door_length)
        return [dest_x1, dest_x2], [dest_y1, dest_y2]

# 获取Polygon对象
def get_polygon(single_poly):
    poly_len = len(single_poly)
    poly = []
    for i in range(poly_len):
        pair = single_poly[i]
        poly.append((pair[key_x], pair[key_y]))
    return Polygon(poly)


# 获取业务空间每个块的最外层轮廓, 组成arr返回, 如果数据不合法发生异常则返回None
def get_outer_polygon_arr(raw_outline):
    try:
        arr = []
        outline_json = json.loads(raw_outline)
        for i in range(len(outline_json)):
            try:
                single_polygon = get_polygon(outline_json[i][0])
            except Exception as ex:
                plpy.info('eroor getting polygon')
                continue
            arr.append(single_polygon)
    except Exception as e:
        return []


# 判断一个点是否在多边形数组中的某一个多边形内
def is_point_in_polygons(point, polygon):
    try:
        return polygon.contains_points([point], None, -0.0001)
    except Exception as e:
        return False


# 检查是否已经加过这个关系(单向检查)
def check_is_in_rel(space_adjacent, probe_id, id):
    if probe_id in space_adjacent:
        rel_set = space_adjacent[probe_id]
        if id in rel_set:
            return True
    return False

# 检查是否关系存在(双向检查)
def check_is_in_rel_bidirection(space_adjacent, probe_id, id):
    is_in = check_is_in_rel(space_adjacent, probe_id, id)
    is_in = is_in or check_is_in_rel(space_adjacent, id, probe_id)
    return is_in

# 将业务空间相邻关系插入map中
def insert_into_rel(space_adjacent, probe_id, id):
    if check_is_in_rel_bidirection(space_adjacent, probe_id, id):
        return
    if probe_id not in space_adjacent:
        space_adjacent[probe_id] = set()
    rel_set = space_adjacent[probe_id]
    rel_set.add(id)

# 计算一个分组内的邻接关系, space_adjacent用来保存关系, 结构是  space_id  -->  set(space_id)
def calc_adjacent_sub_arr(space_sub_arr, space_adjacent, space_info):
    # space_sub_arr 是一个楼层内的一类业务空间
    for space_row in space_sub_arr:
        raw_outline = space_row.get(column_outline)
        id = space_row.get(column_id)
        space_info[id] = space_row
        outline_json = json.loads(raw_outline)
        for i in range(len(outline_json)):
            single_poly = outline_json[i][0]
            poly_len = len(single_poly)
            for idx in range(1, poly_len):
                # 线段
                segment_point1 = single_poly[idx - 1]
                segment_point2 = single_poly[idx]
                # segment_length = get_segment_distance(segment_point1[key_x], segment_point1[key_y], segment_point2[key_x], segment_point2[key_y])
                x_arr, y_arr = get_points(segment_point1[key_x], segment_point1[key_y], segment_point2[key_x], segment_point2[key_y])
                for probe_space_row in space_sub_arr:
                    probe_polygon = probe_space_row.get(column_added_polygon)
                    probe_id = probe_space_row.get(column_id)
                    # 如果自己跟自己比较或者已经存在该关系的话, 跳过
                    if probe_id == id or check_is_in_rel_bidirection(space_adjacent, probe_id, id):
                        continue
                    for arr_index in range(0, x_arr):
                        prob_x = x_arr[arr_index]
                        prob_y = y_arr[arr_index]
                        if is_point_in_polygons((prob_x, prob_y), probe_polygon):
                            insert_into_rel(space_adjacent, probe_id, id)
                            break

# 将输入数据按照楼层id, 业务空间类型分类
def classify(space_list):
    current_floor_id = ''
    current_object_type = ''
    current_sub_arr = []
    space_arr = []
    for row in space_list:
        if row.get(column_floor_id) == current_floor_id and row.get(column_object_type) == current_object_type:
            current_sub_arr.append(row)
        else:
            current_floor_id = row.get(column_floor_id)
            current_object_type = row.get(column_object_type)
            current_sub_arr = [row]
            space_arr.append(current_sub_arr)
        row[column_added_polygon] = get_outer_polygon_arr(row.get(column_outline))
    for sub_arr in space_arr:
        if len(sub_arr) == 1:
            space_arr.remove(sub_arr)
    return space_arr

# 计算业务空间相邻
def calc_space_adjacent(space_list):
    # 给业务空间按照所属楼层和业务空间类型分组
    space_arr = classify(space_list)
    space_adjacent = dict()
    space_info = dict()
    for space_sub_arr in space_arr:
        calc_adjacent_sub_arr(space_sub_arr, space_adjacent, space_info)
    return space_adjacent, space_info


try:
    # 将下面对数据库的操作作为一个事务, 出异常则自动rollback
    with plpy.subtransaction():
        # 获取有所在楼层的所有业务空间, 并按照所在楼层和业务空间类型排序, 方便分组
        sql_str = "SELECT sp.id, sp.project_id, rel.floor_id, sp.object_type, sp.bim_location, sp.outline FROM zone_space_base sp inner join r_sp_in_fl rel on rel.space_id = sp.id where sp.project_id = $1 and rel.project_id = $1 and outline is not null order by floor_id, object_type"
        space_data_plan = plpy.prepare(sql_str, ["text"])
        space_data = space_data_plan.execute([project_id])
        rel_data, space_info = calc_space_adjacent(space_data)
        # 删除以前的业务空间相邻关系
        delete_plan = plpy.prepare("delete from r_spatial_connection where project_id = $1 and sign = 2", ["text"])
        delete_plan.execute([project_id])
        for space_id1, space_id2 in rel_data:
            space1 = space_info[space_id1]
            space2 = space_info[space_id2]
            delete_duplicate_plan = plpy.prepare("delete from r_spatial_connection where space_id_one = $1 and space_id_two = $2", ["text", "text"])
            delete_duplicate_plan.execute([space_id1, space_id2])
            insert_plan = plpy.prepare("insert into r_spatial_connection(project_id, location_one, location_two, space_id_one, space_id_two, sign, graph_type, floor_id, zone_type) values($1, $2, $3, $4, $5, 2, 'SpaceNeighborhood', $6, $7)", ["text", "text", "text", "text", "text", "text", "text"])
            insert_plan.execute([project_id, space1.get(column_bim_location), space2.get(column_bim_location), space_id1, space2, space1.get(column_floor_id), space1.get(column_object_type)])
except Exception as e:
    plpy.warning(e)
    return False
else:
    return True
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100










CREATE OR REPLACE FUNCTION "public"."rel_sy2bd"("project_id" varchar)
  RETURNS "pg_catalog"."bool" AS $BODY$
try:
    # 将下面对数据库的操作作为一个事务, 出异常则自动rollback
    with plpy.subtransaction():
        delete_plan = plpy.prepare("delete from r_sy_in_bd where project_id = $1 and sign = 2", ["text"])
        delete_plan.execute([project_id])
        join_plan = plpy.prepare("select sy.sys_id, eq.building_id from r_sy_eq sy inner join equipment eq on sy.equip_id = eq.id where eq.project_id = $1 and sy.project_id = $1 and eq.building_id is not null", ["text"])
        rel = join_plan.execute([project_id])
        sy2bd = dict()
        for row in rel:
            sys = row['sys_id']
            building = row['building_id']
            if sys not in sy2bd:
                sy2bd[sys] = set()
            sys_set = sy2bd[sys]
            sys_set.add(building)
        for sid, sys_set in sy2bd.items():
            for building in sys_set:
                try:
                    plan = plpy.prepare("insert into r_sy_in_bd(sys_id, building_id, project_id, sign) values($1, $2, $3, 2)", ["text", "text", "text"])
                    plan.execute([sid, building, project_id])
                except Exception as ex:
                    pass
except Exception as e:
    plpy.warning(e)
    return False
else:
    return True
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100







CREATE OR REPLACE FUNCTION "public"."rel_sy2sp"("project_id" varchar)
  RETURNS "pg_catalog"."bool" AS $BODY$
try:
    input_tables = ['r_eq_in_sp_zone_air_conditioning', 'r_eq_in_sp_zone_clean', 'r_eq_in_sp_zone_domestic_water_supply', 'r_eq_in_sp_zone_fire', 'r_eq_in_sp_zone_function',
        'r_eq_in_sp_zone_general', 'r_eq_in_sp_zone_heating', 'r_eq_in_sp_zone_lighting', 'r_eq_in_sp_zone_network', 'r_eq_in_sp_zone_power_supply', 'r_eq_in_sp_zone_security', 'r_eq_in_sp_zone_tenant']
    input_tables = ['relationship.r_eq2sp']
    output_tables = ['r_sy_in_sp_zone_air_conditioning', 'r_sy_in_sp_zone_clean', 'r_sy_in_sp_zone_domestic_water_supply',
        'r_sy_in_sp_zone_fire', 'r_sy_in_sp_zone_function', 'r_sy_in_sp_zone_general', 'r_sy_in_sp_zone_heating', 'r_sy_in_sp_zone_lighting',
        'r_sy_in_sp_zone_network', 'r_sy_in_sp_zone_power_supply', 'r_sy_in_sp_zone_security', 'r_sy_in_sp_zone_tenant']
    output_tables = ['relationship.r_sy2sp']
    # 将下面对数据库的操作作为一个事务, 出异常则自动rollback
    with plpy.subtransaction():
        for i in range(len(output_tables)):
            delete_plan = plpy.prepare("delete from {0} where project_id = $1 and sign = 2 and type = 'sy2sp'".format(output_tables[i]), ["text"])
            delete_plan.execute([project_id])
            join_plan = plpy.prepare("select sy.sys_id, rel.sp_id space_id, zone_type from r_sy_eq sy inner join {0} rel on sy.equip_id = rel.eq_id where rel.project_id = $1 and sy.project_id = $1 and rel.type = 'eq2sp_in'".format(input_tables[i]), ["text"])
            rel = join_plan.execute([project_id])
            sy2sp = dict()
            zone_type_map = dict()
            for row in rel:
                sys = row['sys_id']
                space = row['space_id']
                space_zone = row['zone_type']
                zone_type_map[space] = space_zone
                if sys not in sy2sp:
                    sy2sp[sys] = set()
                space_set = sy2sp[sys]
                space_set.add(space)
            for sid, space_set in sy2sp.items():
                for space in space_set:
                    plan = plpy.prepare("insert into {0}(sy_id, sp_id, project_id, type, zone_type, sign) values($1, $2, $3, 'sy2sp', '{1}', 2)".format(output_tables[i], zone_type_map[space]), ["text", "text", "text"])
                    plan.execute([sid, space, project_id])
except Exception as e:
    plpy.warning(e)
    return False
else:
    return True
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100









CREATE OR REPLACE FUNCTION "public"."rel_sh2bd"("project_id" varchar)
  RETURNS "pg_catalog"."bool" AS $BODY$
try:
    # 将下面对数据库的操作作为一个事务, 出异常则自动rollback
    with plpy.subtransaction():
        delete_plan = plpy.prepare("delete from  r_sh_in_bd rel where rel.project_id = $1 and sign = 2", ["text"])
        delete_plan.execute([project_id])
        join_plan = plpy.prepare("select distinct shsp.sh_id shaft_id, spbd.building_id building_id from relationship.r_sh2sp shsp inner join r_sp_in_bd spbd on shsp.sp_id = spbd.space_id where shsp.project_id = $1 and spbd.project_id = $1", ["text"])
        rel = join_plan.execute([project_id])
        for row in rel:
            shaft_id = row['shaft_id']
            building_id = row['building_id']
            try:
                plan = plpy.prepare("insert into r_sh_in_bd(shaft_id, building_id, project_id, sign) values($1, $2, $3, 2)", ["text", "text", "text"])
                plan.execute([shaft_id, building_id, project_id])
            except Exception as ex:
                pass
except Exception as e:
    plpy.warning(e)
    return False
else:
    return True
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100













create function rel_sp2sp_v2(project_id character varying) returns boolean
    language plpython3u
as
$$
import math
import json
from shapely.geometry import Polygon

column_project_id = 'project_id'
column_location_one = 'location_one'
column_location_two = 'location_two'
column_space_id_one = 'space_id_one'
column_space_id_two = 'space_id_two'
column_zone_type = 'zone_type'
column_added_polygon = 'polygon'

column_id = 'id'
column_bim_location = 'bim_location'
column_floor_id = 'floor_id'
column_object_type = 'object_type'
column_outline = 'outline'
key_x = 'X'
key_y = 'Y'

delta_distance = 200
door_length = 1500


# 获取两点之间的距离
def get_segment_distance(x1, y1, x2, y2):
    x_diff = x1 - x2
    y_diff = y1 - y2
    return math.sqrt(x_diff ** 2 + y_diff ** 2)


# 获取垂直方向的斜率
def get_vertical_k(k):
    if k is None:
        return 0
    if k == 0:
        return None
    return 1 / k

# 根据点 x, y 计算 y = kx + a 中的 a
def get_a_by_point_k(x, y, k):
    if k is None:
        return None
    return y - k * x


# 在直线   y = kx + a 上找到距离点 base_x, base_y 距离为distance 的坐标点(两个坐标点)
def get_point_by_distance(base_x, base_y, k, a, distance):
    if k is None:
        return base_x, base_y + distance, base_x, base_y - distance
    vector_x1 = math.sqrt(distance ** 2 / (1 + k ** 2))
    vector_x2 = -vector_x1
    vector_y1 = k * vector_x1
    vector_y2 = k * vector_x2
    return base_x + vector_x1, base_y + vector_y1, base_x + vector_x2, base_y + vector_y2

# 在直线 y = kx + a 上找到一个点, 该点距离点(base_x, base_y) 长度为 distance, 并且距离vec_x, vec_y最近
def get_point_by_distance_on_segment(base_x, base_y, k, a, distance, vec_x, vec_y):
    x1, y1, x2, y2 = get_point_by_distance(base_x, base_y, k, a, distance)
    distance1 = get_segment_distance(x1, y1, vec_x, vec_y)
    distance2 = get_segment_distance(x2, y2, vec_x, vec_y)
    if distance1 > distance2:
        return x2, y2
    return x1, y1


# 获取输入的两点之间开门之后门的坐标点
# 返回格式
# 如果线段距离小于等于door_length的情况 : x1, y1, x2, y2
# 如果线段距离大于door_length的情况 (两端各开一个门): x1, y1, x2, y2, x3, y3, x4, y4
def get_points(x1, y1, x2, y2):
    # 如果两个点是同一个点
    if x1 == y1 and x2 == y2:
        return None, None, None, None
    # 计算线段的距离
    distance = get_segment_distance(x1, y1, x2, y2)
    # 计算当前线段的 k
    if x1 == x2:
        k = None
    else:
        k = (y1 - y2) / (x1 - x2)
    # 计算垂直方向的k
    vertical_k = get_vertical_k(k)
    # 计算当前线段的 a
    a = get_a_by_point_k(x1, y1, k)
    # 距离大于delta_distance, 则足够开两个门
    if distance > delta_distance:
        seg_x1, seg_y1 = get_point_by_distance_on_segment(x1, y1, k, a, delta_distance, x2, y2)
        seg_x2, seg_y2 = get_point_by_distance_on_segment(x2, y2, k, a, delta_distance, x1, y1)
        vertical_a1 = get_a_by_point_k(seg_x1, seg_y1, vertical_k)
        vertical_a2 = get_a_by_point_k(seg_x2, seg_y2, vertical_k)
        dest_x1, dest_y1, dest_x2, dest_y2 = get_point_by_distance(seg_x1, seg_y1, vertical_k, vertical_a1, door_length)
        dest_x3, dest_y3, dest_x4, dest_y4 = get_point_by_distance(seg_x2, seg_y2, vertical_k, vertical_a2, door_length)
        return [dest_x1, dest_x2, dest_x3, dest_x4], [dest_y1, dest_y2, dest_y3, dest_y4]
    else:
        # 距离太小, 在中间开门
        seg_x1, seg_y1 = get_point_by_distance_on_segment(x1, y1, k, a, distance/2, x2, y2)
        vertical_a1 = get_a_by_point_k(seg_x1, seg_y1, vertical_k)
        dest_x1, dest_y1, dest_x2, dest_y2 = get_point_by_distance(seg_x1, seg_y1, vertical_k, vertical_a1, door_length)
        return [dest_x1, dest_x2], [dest_y1, dest_y2]

# 获取Polygon对象
def get_polygon(single_poly):
    poly_len = len(single_poly)
    poly = []
    for i in range(poly_len):
        pair = single_poly[i]
        poly.append((pair[key_x], pair[key_y]))
        # plpy.info((pair[key_x], pair[key_y]))
    return Polygon(poly)


# 获取业务空间每个块的最外层轮廓, 组成arr返回, 如果数据不合法发生异常则返回None
def get_outer_polygon_arr(raw_outline):
    try:
        arr = []
        outline_json = json.loads(raw_outline)
        for i in range(len(outline_json)):
            try:
                single_polygon = get_polygon(outline_json[i][0])
            except Exception as ex:
                # plpy.info(outline_json[i][0])
                plpy.info('eroor getting polygon')
                continue
            arr.append(single_polygon)
        return arr
    except Exception as e:
        return []


# 判断一个点是否在多边形数组中的某一个多边形内
def is_point_in_polygons(point, polygon_arr):
    try:
        for polygon in polygon_arr:
            try:
                if polygon.contains_points([point], None, -0.0001):
                    return True
            except Exception as ee:
                plpy.warning("point in polygon : {0}".format(ee))
        return False
    except Exception as e:
        plpy.warning(e)
        return False


# 检查是否已经加过这个关系(单向检查)
def check_is_in_rel(space_adjacent, probe_id, id):
    if probe_id in space_adjacent:
        rel_set = space_adjacent[probe_id]
        if id in rel_set:
            return True
    return False

# 检查是否关系存在(双向检查)
def check_is_in_rel_bidirection(space_adjacent, probe_id, id):
    is_in = check_is_in_rel(space_adjacent, probe_id, id)
    is_in = is_in or check_is_in_rel(space_adjacent, id, probe_id)
    return is_in

# 将业务空间相邻关系插入map中
def insert_into_rel(space_adjacent, probe_id, id):
    if check_is_in_rel_bidirection(space_adjacent, probe_id, id):
        return
    if probe_id not in space_adjacent:
        space_adjacent[probe_id] = set()
    rel_set = space_adjacent[probe_id]
    rel_set.add(id)

# 计算一个分组内的邻接关系, space_adjacent用来保存关系, 结构是  space_id  -->  set(space_id)
def calc_adjacent_sub_arr(space_sub_arr, space_adjacent, space_info):
    # space_sub_arr 是一个楼层内的一类业务空间
    for space_row in space_sub_arr:
        raw_outline = space_row.get(column_outline)
        id = space_row.get(column_id)
        space_info[id] = space_row
        outline_json = json.loads(raw_outline)
        #plpy.info("outline_json : {0}".format(len(outline_json)))
        for i in range(len(outline_json)):
            try:
                single_poly = outline_json[i][0]
            except Exception as ee :
                # plpy.warning("error outline : {0}, id : {1}".format(outline_json, id))
                continue
            #single_poly = outline_json[i][0]
            poly_len = len(single_poly)
            for idx in range(1, poly_len):
                # 线段
                segment_point1 = single_poly[idx - 1]
                segment_point2 = single_poly[idx]
                # segment_length = get_segment_distance(segment_point1[key_x], segment_point1[key_y], segment_point2[key_x], segment_point2[key_y])
                x_arr, y_arr = get_points(segment_point1[key_x], segment_point1[key_y], segment_point2[key_x], segment_point2[key_y])
                # plpy.info("x_arr : {0} , y_arr : {1}".format(x_arr, y_arr))
                for probe_space_row in space_sub_arr:
                    probe_polygon_arr = probe_space_row.get(column_added_polygon)
                    probe_id = probe_space_row.get(column_id)
                    # 如果自己跟自己比较或者已经存在该关系的话, 跳过
                    if probe_id == id or check_is_in_rel_bidirection(space_adjacent, probe_id, id):
                        continue
                    for arr_index in range(0, len(x_arr)):
                        prob_x = x_arr[arr_index]
                        prob_y = y_arr[arr_index]
                        is_hit = is_point_in_polygons((prob_x, prob_y), probe_polygon_arr)
                        # plpy.info("is hit : {0}".format(is_hit))
                        if is_hit:
                            plpy.info("hit")
                            insert_into_rel(space_adjacent, probe_id, id)
                            break

# 将输入数据按照楼层id, 业务空间类型分类
def classify(space_list):
    current_floor_id = ''
    current_object_type = ''
    current_sub_arr = []
    space_arr = []
    for row in space_list:
        if row.get(column_floor_id) == current_floor_id and row.get(column_object_type) == current_object_type:
            current_sub_arr.append(row)
        else:
            current_floor_id = row.get(column_floor_id)
            current_object_type = row.get(column_object_type)
            current_sub_arr = [row]
            space_arr.append(current_sub_arr)
        row[column_added_polygon] = get_outer_polygon_arr(row.get(column_outline))
    for sub_arr in space_arr:
        if len(sub_arr) == 1:
            space_arr.remove(sub_arr)
    return space_arr

# 计算业务空间相邻
def calc_space_adjacent(space_list):
    # 给业务空间按照所属楼层和业务空间类型分组
    space_arr = classify(space_list)
    space_adjacent = dict()
    space_info = dict()
    for space_sub_arr in space_arr:
        calc_adjacent_sub_arr(space_sub_arr, space_adjacent, space_info)
    return space_adjacent, space_info


#try:
# 将下面对数据库的操作作为一个事务, 出异常则自动rollback
with plpy.subtransaction():
    # 获取有所在楼层的所有业务空间, 并按照所在楼层和业务空间类型排序, 方便分组
    sql_str = "SELECT sp.id, sp.project_id, rel.floor_id, sp.object_type, sp.bim_location, sp.outline FROM zone_space_base sp inner join r_sp_in_fl rel on rel.space_id = sp.id where sp.project_id = $1 and rel.project_id = $1 and outline is not null order by floor_id, object_type"
    space_data_plan = plpy.prepare(sql_str, ["text"])
    space_data = space_data_plan.execute([project_id])
    plpy.info("space data : {0}".format(len(space_data)))
    rel_data, space_info = calc_space_adjacent(space_data)
    # 删除以前的业务空间相邻关系
    delete_plan = plpy.prepare("delete from r_spatial_connection where project_id = $1 and sign = 2", ["text"])
    delete_plan.execute([project_id])
    plpy.info("rel_data : {0}".format(len(rel_data)))
    for space_id1, space_id2 in rel_data:
        space1 = space_info[space_id1]
        space2 = space_info[space_id2]
        delete_duplicate_plan = plpy.prepare("delete from r_spatial_connection where space_id_one = $1 and space_id_two = $2", ["text", "text"])
        delete_duplicate_plan.execute([space_id1, space_id2])
        insert_plan = plpy.prepare("insert into r_spatial_connection(project_id, location_one, location_two, space_id_one, space_id_two, sign, graph_type, floor_id, zone_type) values($1, $2, $3, $4, $5, 2, 'SpaceNeighborhood', $6, $7)", ["text", "text", "text", "text", "text", "text", "text"])
        insert_plan.execute([project_id, space1.get(column_bim_location), space2.get(column_bim_location), space_id1, space2, space1.get(column_floor_id), space1.get(column_object_type)])
    return True
#except Exception as e:
#    plpy.warning(e)
#    return False
#else:
#    return True
$$;

alter function rel_sp2sp_v2(varchar) owner to postgres;












CREATE OR REPLACE FUNCTION "public"."rel_sp2sp_transport"("project_id" varchar)
  RETURNS "pg_catalog"."bool" AS $BODY$
import math
import json
from shapely.geometry import Polygon

column_project_id = 'project_id'
column_location_one = 'location_one'
column_location_two = 'location_two'
column_space_id_one = 'space_id_one'
column_space_id_two = 'space_id_two'
column_zone_type = 'zone_type'
column_added_polygon = 'polygon'

column_id = 'id'
column_bim_location = 'bim_location'
column_floor_id = 'floor_id'
column_object_type = 'object_type'
column_outline = 'outline'
key_x = 'X'
key_y = 'Y'

delta_distance = 200
door_length = 1500


# 获取两点之间的距离
def get_segment_distance(x1, y1, x2, y2):
    x_diff = x1 - x2
    y_diff = y1 - y2
    return math.sqrt(x_diff ** 2 + y_diff ** 2)


# 获取垂直方向的斜率
def get_vertical_k(k):
    if k is None:
        return 0
    if k == 0:
        return None
    return 1 / k

# 根据点 x, y 计算 y = kx + a 中的 a
def get_a_by_point_k(x, y, k):
    if k is None:
        return None
    return y - k * x


# 在直线   y = kx + a 上找到距离点 base_x, base_y 距离为distance 的坐标点(两个坐标点)
def get_point_by_distance(base_x, base_y, k, a, distance):
    if k is None:
        return base_x, base_y + distance, base_x, base_y - distance
    vector_x1 = math.sqrt(distance ** 2 / (1 + k ** 2))
    vector_x2 = -vector_x1
    vector_y1 = k * vector_x1
    vector_y2 = k * vector_x2
    return base_x + vector_x1, base_y + vector_y1, base_x + vector_x2, base_y + vector_y2

# 在直线 y = kx + a 上找到一个点, 该点距离点(base_x, base_y) 长度为 distance, 并且距离vec_x, vec_y最近
def get_point_by_distance_on_segment(base_x, base_y, k, a, distance, vec_x, vec_y):
    x1, y1, x2, y2 = get_point_by_distance(base_x, base_y, k, a, distance)
    distance1 = get_segment_distance(x1, y1, vec_x, vec_y)
    distance2 = get_segment_distance(x2, y2, vec_x, vec_y)
    if distance1 > distance2:
        return x2, y2
    return x1, y1


# 获取输入的两点之间开门之后门的坐标点
# 返回格式
# 如果线段距离小于等于door_length的情况 : x1, y1, x2, y2
# 如果线段距离大于door_length的情况 (两端各开一个门): x1, y1, x2, y2, x3, y3, x4, y4
def get_points(x1, y1, x2, y2):
    # 如果两个点是同一个点
    if x1 == y1 and x2 == y2:
        return None, None, None, None
    # 计算线段的距离
    distance = get_segment_distance(x1, y1, x2, y2)
    # 计算当前线段的 k
    if x1 == x2:
        k = None
    else:
        k = (y1 - y2) / (x1 - x2)
    # 计算垂直方向的k
    vertical_k = get_vertical_k(k)
    # 计算当前线段的 a
    a = get_a_by_point_k(x1, y1, k)
    # 距离大于delta_distance, 则足够开两个门
    if distance > delta_distance:
        seg_x1, seg_y1 = get_point_by_distance_on_segment(x1, y1, k, a, delta_distance, x2, y2)
        seg_x2, seg_y2 = get_point_by_distance_on_segment(x2, y2, k, a, delta_distance, x1, y1)
        vertical_a1 = get_a_by_point_k(seg_x1, seg_y1, vertical_k)
        vertical_a2 = get_a_by_point_k(seg_x2, seg_y2, vertical_k)
        dest_x1, dest_y1, dest_x2, dest_y2 = get_point_by_distance(seg_x1, seg_y1, vertical_k, vertical_a1, door_length)
        dest_x3, dest_y3, dest_x4, dest_y4 = get_point_by_distance(seg_x2, seg_y2, vertical_k, vertical_a2, door_length)
        return [dest_x1, dest_x2, dest_x3, dest_x4], [dest_y1, dest_y2, dest_y3, dest_y4]
    else:
        # 距离太小, 在中间开门
        seg_x1, seg_y1 = get_point_by_distance_on_segment(x1, y1, k, a, distance/2, x2, y2)
        vertical_a1 = get_a_by_point_k(seg_x1, seg_y1, vertical_k)
        dest_x1, dest_y1, dest_x2, dest_y2 = get_point_by_distance(seg_x1, seg_y1, vertical_k, vertical_a1, door_length)
        return [dest_x1, dest_x2], [dest_y1, dest_y2]

# 获取Polygon对象
def get_polygon(single_poly):
    poly_len = len(single_poly)
    poly = []
    for i in range(poly_len):
        pair = single_poly[i]
        poly.append((pair[key_x], pair[key_y]))
        # plpy.info((pair[key_x], pair[key_y]))
    return Polygon(poly)


# 获取业务空间每个块的最外层轮廓, 组成arr返回, 如果数据不合法发生异常则返回None
def get_outer_polygon_arr(raw_outline):
    try:
        arr = []
        outline_json = json.loads(raw_outline)
        for i in range(len(outline_json)):
            try:
                single_polygon = get_polygon(outline_json[i][0])
            except Exception as ex:
                # plpy.info(outline_json[i][0])
                plpy.info('eroor getting polygon')
                continue
            arr.append(single_polygon)
        return arr
    except Exception as e:
        return []


# 判断一个点是否在多边形数组中的某一个多边形内
def is_point_in_polygons(point, polygon_arr):
    try:
        for polygon in polygon_arr:
            try:
                if polygon.contains_points([point], None, -0.0001):
                    return True
            except Exception as ee:
                plpy.warning("point in polygon : {0}".format(ee))
        return False
    except Exception as e:
        plpy.warning(e)
        return False


# 检查是否已经加过这个关系(单向检查)
def check_is_in_rel(space_adjacent, probe_id, id):
    if probe_id in space_adjacent:
        rel_set = space_adjacent[probe_id]
        if id in rel_set:
            return True
    return False

# 检查是否关系存在(双向检查)
def check_is_in_rel_bidirection(space_adjacent, probe_id, id):
    is_in = check_is_in_rel(space_adjacent, probe_id, id)
    is_in = is_in or check_is_in_rel(space_adjacent, id, probe_id)
    return is_in

# 将业务空间相邻关系插入map中
def insert_into_rel(space_adjacent, probe_id, id):
    if check_is_in_rel_bidirection(space_adjacent, probe_id, id):
        return
    if probe_id not in space_adjacent:
        space_adjacent[probe_id] = set()
    rel_set = space_adjacent[probe_id]
    rel_set.add(id)

# 计算一个分组内的邻接关系, space_adjacent用来保存关系, 结构是  space_id  -->  set(space_id)
def calc_adjacent_sub_arr(space_sub_arr, space_adjacent, space_info):
    # space_sub_arr 是一个楼层内的一类业务空间
    for space_row in space_sub_arr:
        raw_outline = space_row.get(column_outline)
        id = space_row.get(column_id)
        space_info[id] = space_row
        outline_json = json.loads(raw_outline)
        #plpy.info("outline_json : {0}".format(len(outline_json)))
        for i in range(len(outline_json)):
            try:
                single_poly = outline_json[i][0]
            except Exception as ee :
                # plpy.warning("error outline : {0}, id : {1}".format(outline_json, id))
                continue
            #single_poly = outline_json[i][0]
            poly_len = len(single_poly)
            for idx in range(1, poly_len):
                # 线段
                segment_point1 = single_poly[idx - 1]
                segment_point2 = single_poly[idx]
                # segment_length = get_segment_distance(segment_point1[key_x], segment_point1[key_y], segment_point2[key_x], segment_point2[key_y])
                x_arr, y_arr = get_points(segment_point1[key_x], segment_point1[key_y], segment_point2[key_x], segment_point2[key_y])
                # plpy.info("x_arr : {0} , y_arr : {1}".format(x_arr, y_arr))
                for probe_space_row in space_sub_arr:
                    probe_polygon_arr = probe_space_row.get(column_added_polygon)
                    probe_id = probe_space_row.get(column_id)
                    # 如果自己跟自己比较或者已经存在该关系的话, 跳过
                    if probe_id == id or check_is_in_rel_bidirection(space_adjacent, probe_id, id):
                        continue
                    for arr_index in range(0, len(x_arr)):
                        prob_x = x_arr[arr_index]
                        prob_y = y_arr[arr_index]
                        is_hit = is_point_in_polygons((prob_x, prob_y), probe_polygon_arr)
                        # plpy.info("is hit : {0}".format(is_hit))
                        if is_hit:
                            plpy.info("hit")
                            insert_into_rel(space_adjacent, probe_id, id)
                            break

# 将输入数据按照楼层id, 业务空间类型分类
def classify(space_list):
    current_floor_id = ''
    current_object_type = ''
    current_sub_arr = []
    space_arr = []
    for row in space_list:
        if row.get(column_floor_id) == current_floor_id and row.get(column_object_type) == current_object_type:
            current_sub_arr.append(row)
        else:
            current_floor_id = row.get(column_floor_id)
            current_object_type = row.get(column_object_type)
            current_sub_arr = [row]
            space_arr.append(current_sub_arr)
        row[column_added_polygon] = get_outer_polygon_arr(row.get(column_outline))
    for sub_arr in space_arr:
        if len(sub_arr) == 1:
            space_arr.remove(sub_arr)
    return space_arr

# 计算业务空间相邻
def calc_space_adjacent(space_list):
    # 给业务空间按照所属楼层和业务空间类型分组
    space_arr = classify(space_list)
    space_adjacent = dict()
    space_info = dict()
    for space_sub_arr in space_arr:
        calc_adjacent_sub_arr(space_sub_arr, space_adjacent, space_info)
    return space_adjacent, space_info


#try:
# 将下面对数据库的操作作为一个事务, 出异常则自动rollback
with plpy.subtransaction():
    # 获取有所在楼层的所有业务空间, 并按照所在楼层和业务空间类型排序, 方便分组
    sql_str = "SELECT sp.id, rel.floor_id, sp.object_type, sp.bim_location, sp.outline FROM zone_space_base sp inner join r_sp_in_fl rel on rel.space_id = sp.id where sp.project_id = $1 and rel.project_id = $1 and outline is not null order by floor_id, object_type"
    space_data_plan = plpy.prepare(sql_str, ["text"])
    space_data = space_data_plan.execute([project_id])
    plpy.info("space data : {0}".format(len(space_data)))
    rel_data, space_info = calc_space_adjacent(space_data)
    # 删除以前的业务空间相邻关系
    delete_plan = plpy.prepare("delete from r_spatial_connection where project_id = $1 and sign = 2", ["text"])
    delete_plan.execute([project_id])
    plpy.info("rel_data : {0}".format(len(rel_data)))
    for space_id1, space_id2 in rel_data:
        space1 = space_info[space_id1]
        space2 = space_info[space_id2]
        delete_duplicate_plan = plpy.prepare("delete from r_spatial_connection where space_id_one = $1 and space_id_two = $2", ["text", "text"])
        delete_duplicate_plan.execute([space_id1, space_id2])
        insert_plan = plpy.prepare("insert into r_spatial_connection(project_id, location_one, location_two, space_id_one, space_id_two, sign, graph_type, floor_id, zone_type) values($1, $2, $3, $4, $5, 2, 'SpaceNeighborhood', $6, $7)", ["text", "text", "text", "text", "text", "text", "text"])
        insert_plan.execute([project_id, space1.get(column_bim_location), space2.get(column_bim_location), space_id1, space2, space1.get(column_floor_id), space1.get(column_object_type)])
    return True
#except Exception as e:
#    plpy.warning(e)
#    return False
#else:
#    return True
$BODY$
  LANGUAGE plpython3u VOLATILE
  COST 100

































